#!/usr/bin/env bash

#
# coup
#
# Copyright (c) 2014-2016, rnlf
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of the MIT license. See LICENSE for details.
#


function pushd() {
  command pushd "$@" > /dev/null
}

function popd() {
  command popd "$@" > /dev/null
}


source "coup.conf"
source "$COUP_DIR"/common.sh

function build_spritesheets() {
  for i in $(find ./data_src/sprites -name '*.ase' -or -name '*.aseprite'); do
    dir=$(dirname "$i" | sed -e 's:\./data_src:data:')
    fname=$(basename "$i")

    jsonfile=$(echo "$fname" | sed -e 's:\.ase$:.json:')
    pngfile=$(echo "$fname" | sed -e 's:\.ase$:.png:')

    mkdir -p ".build/$dir"

    # Running aseprite from within the .build directory will give us the correct
    # path for the image filename in the JSON output
    pushd .build
    aseprite -b --border-padding 1 --shape-padding 1 --trim --sheet-pack --list-tags --format json-array --sheet "$dir/$pngfile" --data "$dir/$jsonfile" ../"$i"
    popd
  done
}


function build_tilemaps() {
  for i in $(find ./data_src/tilemaps -name '*.tmx'); do
    dir=$(dirname "$i" | sed -e 's:\./data_src::')
    fname=$(basename "$i")

    luafile=$(echo "$fname" | sed -e 's:\.tmx$:.lua:')

    mkdir -p ".build/data/$dir"

    tiled --export-map "Lua files (*.lua)" "./data_src/$dir/$fname" ".build/data/$dir/$luafile"

    # Fix tileset file paths
    sed -i -e 's:../../../data:data:' ".build/data/$dir/$luafile"

  done
}


function build() {
  if [ "$1" == "-c" ]; then
    echo "Cleaning build directory"
    rm -rf .build
  fi

  mkdir -p .build

  RSYNC_OPTS='--exclude=.git --exclude=.*.swp'

  rsync -r $RSYNC_OPTS src/* .build
  rsync -r $RSYNC_OPTS "$COUP_DIR"/coup .build
  rsync -r $RSYNC_OPTS data .build

  build_spritesheets
  build_tilemaps
}


function download_zipfile() {
  FILENAME=".download/$2"
  BASE_URL="https://bitbucket.org/rude/love/downloads"
  if [ -f "$FILENAME" ]; then
    if [ "$(sha1sum "$FILENAME" | awk '{print $1}')" == "$1" ]; then
      return
    fi
    echo "$2 present but wrong checksum, downloading again."
    rm -f "$FILENAME"
  fi
  echo "Downloading $2..."

  wget -P .download "$BASE_URL"/"$2"
}


function build_win_release() {
  WIN="$1"
  echo "Building $WIN release..."
  BASENAME="love-$COUP_LOVE_VERSION-$WIN"
  PROJNAME="$PROJECT_NAME-$WIN"
  rm -rf .tmp
  mkdir .tmp

  pushd .tmp
  unzip -q ../.download/$BASENAME.zip
  rm -f $BASENAME/{lovec.exe,readme.txt,changes.txt}
  cat $BASENAME/love.exe ../release/"$PROJECT_NAME".love >$BASENAME/"$PROJNAME".exe
  rm -f $BASENAME/love.exe
  mv $BASENAME "$PROJNAME"
  zip -q -r ../release/"$PROJNAME".zip "$PROJNAME"
  rm -rf .tmp
  popd
}


function build_mac_release() {
  if [ -z "$OSX_BUNDLE_IDENTIFIER" ]; then
    echo "To enable building OSX app, set OSX_BUNDLE_IDENTIFIER in coup.conf"
    return
  fi

  if ! which xml >/dev/null; then
    echo "Can't find xmlstarlet. Needed for building OSX app"
    return
  fi

  if ! which lua >/dev/null; then
    echo "Can't find Lua executable. Needed for building OSX app"
    return
  fi

  TITLE=$(cd src/; lua -e 'love = {} require("conf") t = {window={title=""}} love.conf(t) print(t.window.title)')

  if [ -z "$TITLE" ]; then
    echo "conf.lua doesn't set t.window.title. Need that to figure out the name of the OSX app"
    return
  fi

  echo "Building OSX version..."

  BASENAME="love-$COUP_LOVE_VERSION-macosx-x64"
  PROJNAME="$PROJECT_NAME-osx"

  rm -rf .tmp
  mkdir .tmp
  pushd .tmp

  unzip -q ../.download/$BASENAME.zip

  cp ../release/"$PROJECT_NAME".love love.app/Contents/Resources
  
  sed -i -e '/<key>CFBundleIdentifier<\/key>/,+1 s/org\.love2d\.love/'$OSX_BUNDLE_IDENTIFIER'/
  /<key>CFBundleName<\/key>/,+1 s/LÖVE/'$TITLE'/' love.app/Contents/Info.plist

  xml ed -L -d '//dict/key[text()="UTExportedTypeDeclarations"]|//dict/key[text()="UTExportedTypeDeclarations"]/following-sibling::array' love.app/Contents//Info.plist 

  mv love.app "$PROJECT_NAME".app
  zip -q -r ../release/"$PROJECT_NAME-osx.zip" "$PROJECT_NAME".app

  popd
  rm -rf .tmp
}


function download_love_zipfiles() {
  while read -r line; do
    download_zipfile $line
  done < "$COUP_DIR"/downloads.txt
}


function package() {
  if [ -z "$PROJECT_NAME" ]; then
    echo "Please set PROJECT_NAME in coup.conf first."
    exit 1
  fi

  mkdir -p release

  pushd .build
  rm -f ../release/"$PROJECT_NAME".love
  zip -q -r ../release/"$PROJECT_NAME".love .
  popd

  download_love_zipfiles

  build_win_release win32
  build_win_release win64

  build_mac_release
}


if [ -z "$1" ]; then
  VERB="run"
else
  VERB="$1"
fi


case "$VERB" in
  run)
    build
    cd .build
    "$LOVE" .
  ;;
  build)
    build -c
    package
  ;;
  clean)
    rm -rf .build .tmp release .download
esac

