--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

local SpatialHashtable = {}
SpatialHashtable.__index = SpatialHashtable


function SpatialHashtable:iterBinsAABB(bx1, by1, bx2, by2)
  return coroutine.wrap(function()
    local b = self.bins
    local s = self.currentSize

    for y = math.max(s[2], by1), math.min(s[4], by2) do
      local by = b[y]
      for x = math.max(s[1], bx1), math.min(s[3], bx2) do
        coroutine.yield(by[x])
      end
    end
  end)
end


function SpatialHashtable:iterBinsDDA(bx1, by1, bx2, by2)
  return coroutine.wrap(function()
    local b = self.bins
    for x, y in coup.raster.iterLineDDA(bx1, by1, bx2, by2) do
      local by = b[y]
      if by then
        local bx = by[x]
        if bx then
          coroutine.yield(bx)
        end
      end
    end
  end)
end


function SpatialHashtable:collectBinsAABB(bx1, by1, bx2, by2)
  local ret = {}

  for b in self:iterBinsAABB(bx1, by1, bx2, by2) do
    ret[#ret+1] = b
  end

  return ret
end


function SpatialHashtable:iterEntitiesAABB(x1, y1, x2, y2)
  local bx1, by1 = self:hash(x1, y1)
  local bx2, by2 = self:hash(x2, y2)

  return coroutine.wrap(function()
    local seenEnts = {}
    for b in self:iterBinsAABB(bx1, by1, bx2, by2) do
      for k, _ in pairs(b) do
        if not seenEnts[k] then
          coroutine.yield(k)
          seenEnts[k] = k
        end
      end
    end
  end)
end


function SpatialHashtable:findEntitiesInRange(x1, y1, x2, y2)
  local ret = {}
  
  self:foreachEntityInRange(x1, y1, x2, y2, function(k)
    ret[#ret+1] = k
  end)

  return ret
end


function SpatialHashtable:findNeighbours(e)
  local ed = self.knownEntities[e]

  local neighbours = {}

  for i, b in ipairs(ed.bins) do
    for k, j in pairs(b) do
      neighbours[k] = k
    end
  end

  return neighbours
end


function SpatialHashtable:ensureSize(bx1, by1, bx2, by2)
  local s = self.currentSize
  local b = self.bins

  -- create bins above current if necessary
  for y = by1, s[2] - 1 do
    local n = {}
    for x = s[1], s[3] do
      n[x] = {}
    end
    b[y] = n
  end
  s[2] = math.min(s[2], by1)

  -- create bins below current if necessary
  for y = s[4] + 1, by2 do
    local n = {}
    for x = s[1], s[3] do
      n[x] = {}
    end
    b[y] = n
  end
  s[4] = math.max(s[4], by2)

  -- create bins left of current if necessary
  for x = bx1, s[1] - 1 do
    for y = s[2], s[4] do
      b[y][x] = {}
    end
  end
  s[1] = math.min(s[1], bx1)

  -- create bins right of current if necessary
  for x = s[3] + 1, bx2 do
    for y = s[2], s[4] do
      b[y][x] = {}
    end
  end
  s[3] = math.max(s[3], bx2)
end


function SpatialHashtable:hash(x, y)
  local bs = 1.0 / self.binSize
  return math.floor(x * bs), math.floor(y * bs)
end


-- TODO: Basically the same code as insertLine
function SpatialHashtable:insertAABB(e, x1, y1, x2, y2)
  local bx1, by1 = self:hash(x1, y1)
  local bx2, by2 = self:hash(x2, y2)

  local kE = self.knownEntities
  local old = kE[e]
  if old and old.extent then
    if old.extent[1] == bx1 and old.extent[2] == by1 and
       old.extent[3] == bx2 and old.extent[4] == by2 then
      return
    end
    self:remove(e)
  end

  self:ensureSize(bx1, by1, bx2, by2)
  local bins = {}
  local ed = {
    entity = e,
    bins = bins,
    extent = {bx1, by1, bx2, by2}
  }
  for b in self:iterBinsAABB(bx1, by1, bx2, by2) do
    bins[#bins+1] = b
    b[e] = ed
  end
  
  kE[e] = ed
end


function SpatialHashtable:insertLine(e, x1, y1, x2, y2)
  local bx1, by1 = self:hash(x1, y1)
  local bx2, by2 = self:hash(x2, y2)

  local kE = self.knownEntities
  local old = kE[e]
  if old and old.extentDDA then
    if old.extentDDA[1] == bx1 and old.extentDDA[2] == by1 and
       old.extentDDA[3] == bx2 and old.extentDDA[4] == by2 then
      return
    end
    self:remove(e)
  end

  self:ensureSize(bx1, by1, bx2, by2)
  local bins = {}
  local ed = {
    entity = e,
    bins = bins,
    extentDDA = {bx1, by1, bx2, by2}
  }
  for b in self:iterBinsDDA(bx1, by1, bx2, by2) do
    bins[#bins+1] = b
    b[e] = ed
  end
  
  kE[e] = ed
end


function SpatialHashtable:remove(e)
  local kE = self.knownEntities
  local ed = kE[e]

  for i, b in ipairs(ed.bins) do
    b[e] = nil
  end

  kE[e] = nil
end


local function newSpatialHashtable(binSize)
  return setmetatable({
    binSize = binSize,
    currentSize = {1, 1, 1, 1},
    knownEntities = {},
    bins = {{{}}}
  }, SpatialHashtable)
end


return { SpatialHashtable = newSpatialHashtable }
