--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

local BASE = (...):match("(.-)[^%.]+$")
local vector = require(BASE .. ".vector")

local AnimationRepository = {}
AnimationRepository.__index = AnimationRepository

local Sprite = {}
Sprite.__index = Sprite

local Animation = {}
Animation.__index = Animation

function Animation:addFrame(img, x, y, w, h, ox, oy, delay)
  -- TODO memoize Quads
  local iw, ih = img:getDimensions()
  local start = 0
  if #self.frames > 0 then
    start = self.frames[#self.frames].start + self.frames[#self.frames].delay
  end
  self.frames[#self.frames+1] = {
    img = img,
    quad = love.graphics.newQuad(x, y, w, h, iw, ih),
    origin = vector.Vector(ox, oy),
    start = start,
    delay = delay,
    index = #self.frames+1
  }
end


function Animation:getFrame(idx)
  return self.frames[idx+1]
end


function Animation:getFrameForTime(time)
  for i, f in ipairs(self.frames) do
    if time >= f.start and time < f.start + f.delay then
      return i - 1
    end
  end
end


function Animation:frameCount()
  return #self.frames
end


function Animation:totalTime()
  local f = self.frames[#self.frames]
  return f.start + f.delay
end


function AnimationRepository:addAnimation(name)
  local newAnim = setmetatable({frames = {}}, Animation)
  newAnim.name = name

  self.animations[name] = newAnim
  return newAnim
end


function AnimationRepository:getAnimation(name)
  return self.animations[name]
end


function AnimationRepository:newSprite(anim)
  local sprite = setmetatable({repo = self, speed=1.0, time=0.0}, Sprite)
  if anim then
    sprite:play(anim)
  end

  return sprite
end


local function newAnimationRepository()
  return setmetatable({animations = {}}, AnimationRepository)
end


function Sprite:getRepository()
  return self.repo
end


function Sprite:getCurrentFrame()
  if not self.anim then
    return nil
  end

  return self.anim:getFrame(self.anim:getFrameForTime(self.time))
end

-- :chain() and :onEnd() must be called after :play(). :play() resets them.
function Sprite:play(anim, no_rewind)
  self.anim = self.repo:getAnimation(anim)
  if no_rewind then
    self.time = self.time % self.anim:totalTime()
  else
    self.time = 0
  end

  self.nextAnim = nil
  self.endEvent = nil

  return self
end


function Sprite:chain(anim)
  self.nextAnim = anim
  self.endEvent = nil
end


function Sprite:update(dt)
  local t = self.time + self.speed * dt
  local tt = self.anim:totalTime()

  local endEventCount = 0
  if t >= tt then
    endEventCount = math.floor(t / tt)
    if self.nextAnim then
      -- let play() handle the time overflow
      self.time = t
      self:play(self.nextAnim, true)
      return
    end
    t = t % tt
  end

  self.time = t

  if self.endEvent then
    for i = 1, endEventCount do
      -- make sure we fire the newly installed event handler
      local event = self.endEvent
      self.endEvent = nil
      if event then
        event(self)
      end
    end
  end
end


function Sprite:onEnd(func, continuous)
  self.nextAnim = nil
  if continuous then
    local retrigger
    retrigger = function(spr)
      func(spr)
      spr.endEvent = retrigger
    end
    self.endEvent = retrigger
  else
    self.endEvent = func
  end
end


function Sprite:setSpeed(factor)
  self.speed = factor
end

-- TODO add frame-tags for Animations and add events to Sprites when a tag is hit.

return {
  AnimationRepository = newAnimationRepository,
}
