--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

local tables = {}

-- Shallow merge tables from left to right:
-- Keys present in multiple argument tables will be set to the value in the
-- left-most table that has this key.
-- The argument tables are not altered, the merged result is returned
tables.merge = function(...)
  local res = {}
  for _, t in ipairs({...}) do
    for k, v in pairs(t) do
      res[k] = v
    end
  end

  return res
end


-- Set value by name in a table of tables:
-- scoped_set(table, "x.y.z", 10) sets table.x.y.z to 10
-- Useful for reading configuration files or entity properties from map files
-- All tables in the chain must already exist.
-- If create is set, missing tables will be created.
-- Returns true if the value was set correctly, false if not.
function tables.scoped_set(table, key, value, create)
  -- Generate list of scope parts
  local parts = {}
  for i in key:gmatch('[^.]+') do
    parts[#parts+1] = i
  end

  -- Iterate table to find final table
  for i = 1, #parts-1 do
    local nt = table[parts[i]]

    -- Generate intermediate table if required
    if (nt == nil) and create then
      nt = {}
      table[parts[i]] = nt
    end
    table = nt

    if type(table) ~= "table" then
      return false
    end
  end

  table[parts[#parts]] = value
  return true
end


-- Get value by name in a table of tables:
-- scoped_get(table, "x.y.z") returns table.x.y.z
-- Useful for reading configuration files or entity properties from map files
-- All tables in the chain must exist.
function tables.scoped_get(table, key)
  local t2 = table
  for i in key:gmatch('[^.]+') do
    if table == nil then
      return
    end
    table = table[i]
  end
  if t2 == table then return nil end
  return table
end


-- Collect all key:value pairs returned by iterator into a table and
-- return that table. This is the inverse operation of pairs.
function tables.collect(f, a, k)
  local res = {}
  for k, v in f, a, k do
    res[k] = v
  end
  return res
end


-- Collect all values returned by iterator into an array and
-- return that array. This is the inverse operation of pairs.
function tables.array(f, a, k)
  local res = {}
  for _, v in f, a, k do
    res[#res+1] = v
  end
  return res
end


-- Adapt the passed in iterator to to not return the values, but the values as
-- passed through the function func. Example:
-- for k, v in map_iter(coup.math.sq, ipairs(input)) do .... end
--   iterates over the array input and returns the value squared.
function tables.map_iter(func, f, a, k)
  local iterfunc = function(b, l)
    local nk, v = f(b, l)
    if not nk then
      return nil
    end
    return nk, func(v)
  end
  return iterfunc, a, k
end


-- Return a table containing all keys of input with the associated values
-- modified by func.
-- Only use the array part of input if asArray is true.
function tables.map(func, input)
  return tables.collect(tables.map_iter(func, pairs(input)))
end


-- Return a table containing all keys of input with the associated values
-- modified by func.
-- Only use the array part of input if asArray is true.
function tables.map_array(func, input)
  return tables.array(tables.map_iter(func, ipairs(input)))
end


-- Left-associatively reduce the values passed in through the iterator.
function tables.reduce_iter(func, init, f, a, k)
  local acc
  if init then
    acc = init
  else
    k, acc = f(a, k)
  end

  for i, v in f, a, k do
    acc = func(acc, v)
  end

  return acc
end


-- Left-associatively reduce the values in the array.
function tables.reduce(func, input, init)
  return tables.reduce_iter(func, init, ipairs(input))
end


-- Return only those elements in the sequence for which pred returns true.
function tables.filter_iter(pred, f, a, k)
  local iterfunc = function(b, l)
    local v
    l, v = f(b, l)
    while l ~= nil and not pred(v) do
      l, v = f(b, l)
    end
    return l, v
  end
  return iterfunc, a, k
end


-- Return only those elements in the array for which pred returns true.
function tables.filter_array(pred, array)
  return tables.array(tables.filter_iter(pred, ipairs(array)))
end


-- Return only those elements in the table for which pred returns true.
function tables.filter(pred, input)
  return tables.collect(tables.filter_iter(pred, pairs(input)))
end


-- Append the given values to the array
function tables.append(array, ...)
  for i, k in ipairs({...}) do
    array[#array+1] = k
  end
end


-- Concatenate the passed-in arrays and return the result
function tables.concat(...)
  local result = {}
  for i, v in ipairs({...}) do
    for j, w in ipairs(v) do
      result[#result+1] = w
    end
  end

  return result
end


-- Concatenate all passed-in arrays to the first one in-place (changing the first array)
function tables.append_all(array, ...)
  for i, a in ipairs({...}) do
    for i, k in ipairs(a) do
      array[#array+1] = k
    end
  end
end


-- Returns the first value for which pred is true
function tables.first_iter(pred, f, a, k)
  for k, v in f, a, k do
    if pred(v) then
      return k, v
    end
  end
end


-- Returns the first value in the array for which pred is true
function tables.first(pred, array)
  return tables.first_iter(pred, ipairs(array))
end


-- Returns the list of keys of the passed table
function tables.keys(tab)
  local keys = {}
  for k, v in pairs(tab) do
    table.insert(keys, k)
  end
  return keys
end


-- Returns a random element of the passed in array. If rng is nil, love.math.random is used
function tables.random_table(tab, rng)
  local keys = tables.keys(tab)

  local key = tables.random_array(keys, rng)
  return tab[key], key
end


function tables.random_array(arr, rng)
  local idx = 1

  if rng then
    idx = rng:random(1, #arr)
  else
    idx = love.math.random(1, #arr)
  end

  return arr[idx], idx
end

-- Ignores tables in the keys
-- If tables are used as keys, only the same table will be considered equal
function tables.deep_equal(a, b)
  local visited = {}
  local function recurse(a, b)
    if type(a) ~= type(b) then return false end
    if type(a) ~= "table" then return a == b end
    if visited[a] then return visited[a] == b end

    visited[a] = b

    for k, v in pairs(a) do
      if not recurse(v, b[k]) then
        visited[a] = nil
        return false
      end
    end

    for k, v in pairs(b) do
      if not recurse(v, a[k]) then
        visited[a] = nil
        return false
      end
    end

    visited[a] = nil

    return true
  end

  return recurse(a, b)
end


function tables.array_find(array, val)
  for i, v in ipairs(array) do
    if v == val then return i end
  end
  return nil
end


function tables.array_find_f(array, func)
  for i, v in ipairs(array) do
    if func(v) then return i end
  end
  return nil
end


-- TODO: Tests
function tables.range(a, b, i)
  local r = {}
  for j = a, b, (i or 1) do
    r[#r+1] = j
  end
  return r
end


-- TODO: Tests
function tables.reverse_array(a)
  local r = {}
  local n = #a
  for j, i in ipairs(a) do
    r[n-j+1] = i
  end
  return r
end


return tables
