--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

local BASE = (...):match("(.-)[^%.]+$")
local smath = require(BASE .. ".smath")
local tables = require(BASE .. ".tables")

local tween = {}

local createTween

local Tween = {
  -- Set easing mode
  ease = function(self, mode)
    self._interpolate = smath.interpolate_func(mode)
    return self
  end,

  -- Set delay
  delay = function(self, time)
    self._delay = time
    return self
  end,

  -- Set start callback
  on_start = function(self, ...)
    tables.append(self._on_start, ...)
    return self
  end,

  -- Set update callback
  on_update = function(self, ...)
    tables.append(self._on_update, ...)
    return self
  end,

  -- Set complete callback
  on_complete = function(self, ...)
    tables.append(self._on_complete, ...)
    return self
  end,

  -- Chain new tween after this one.
  -- All future setters will change the new tween instead of the current
  after = function(self, obj, time, vars)
    if vars then
      self._next = createTween(self._group, obj, time, vars)
    else
      self._next = createTween(self._group, self._obj, obj, time)
    end
    return self._next
  end,

  -- Abort current tween. No more callback will be called
  abort = function(self)
    self._group:remove(self)
    return self
  end
}

Tween.__index = Tween


createTween = function(group, obj, time, vars)
  local nt = {
    _on_start    = {},
    _on_update   = {},
    _on_complete = {},
    _vars        = {},
    _time        = time,
    _obj         = obj,
    _delay       = 0,
    _curtime     = 0,
    _next        = nil,
    _group       = group,
    _interpolate = smath.interpolate_func('linear')
  }

  for k, v in pairs(vars) do
    nt._vars[#nt._vars+1] = {
      name  = k,
      start = obj[k],
      dest  = v
    }
  end
  

  setmetatable(nt, Tween)
  return nt
end

local Group = {
  -- Create a new tween for object obj to morph all vars over a time seconds
  -- to the new values
  tween = function(self, obj, time, vars)
    local newTween = createTween(self, obj, time, vars)
    self.tweens[#self.tweens+1] = newTween
    return newTween
  end,

  -- Progress all tweens by dt seconds
  update = function(self, dt)
    -- Iterate in reverse
    local i = #self.tweens
    local dtt = dt
    while i >= 1 do
      local t = self.tweens[i]

      -- Tween has delay, subtract dt
      if t._delay > 0 then
        t._delay = t._delay - dt

        -- delay expired, fire on_start of current tween and 
        -- decrease delta time by remaining delay
        if t._delay <= 0 then
          for _, stf in ipairs(t._on_start) do
            stf()
          end
          
          dtt = -t._delay
          t._delay = 0
        end
      end

      -- Tween does not have delay (anymore)
      if t._delay == 0 
          and dtt > 0 then
        t._curtime = t._curtime + dtt
        local alpha = math.min(1.0, t._curtime / t._time)

        -- First update tween vars
        for _, var in ipairs(t._vars) do
          local val = t._interpolate(var.start, var.dest, alpha)
          t._obj[var.name] = val
        end

        -- Then fire on_update functions
        for _, upf in ipairs(t._on_update) do
          upf()
        end

        -- If tween has finished, either replace it with its
        -- _next tween or remove it
        if t._curtime >= t._time then
          -- Fire on_complete functions
          for _, cof in ipairs(t._on_complete) do
            cof()
          end

          if t._next then
            -- Replace tween with _next and set _start to current value
            -- (was just updated a few lines up!)
            self.tweens[i] = t._next
            for _, var in ipairs(t._next._vars) do
              var.start = t._next._obj[var.name]
            end

            -- replace current tween with _next and initialize its
            -- elapsed timer to the overflow from this tween and
            -- subtract current dt: We just redo the current index
            -- and add dt back in again in the next iteration
            t._next._curtime = t._curtime - t._time - dt
            i = i + 1
          else
            -- replace current tween with last tween in the list
            self:remove_idx(i)
          end
        end
      end
      i = i - 1
    end
  end,

  abort_all = function(self)
    for i = 1, #self.tweens do
      self:remove_idx(i)
    end
  end,

  -- Remove tween with index i from group
  remove_idx = function(self, i)
    self.tweens[i] = self.tweens[#self.tweens]
    self.tweens[#self.tweens] = nil
  end,

  -- Remove tween from group
  remove = function(self, tween)
    local i = tables.first(function(t) return t == tween end, self.tweens)
    self:remove_idx(i)
  end
}

Group.__index = Group

-- Create new tween group
tween.group = function()
  return setmetatable({tweens = {}}, Group)
end


return tween
