--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

local BASE = (...):match("(.-)[^%.]+$")
local smath = require(BASE .. ".smath")

local Vector = {}
Vector.__index = Vector


local newVector = function(x, y, z)
  return setmetatable(
    {
      x = x or 0,
      y = y or 0,
      z = z or 0,
    }, Vector)
end


function Vector.__add(a, b)
  return setmetatable(
    {
      x = a.x + b.x,
      y = a.y + b.y,
      z = a.z + b.z,
    }, Vector)
end


function Vector.__sub(a, b)
  return setmetatable(
    {
      x = a.x - b.x,
      y = a.y - b.y,
      z = a.z - b.z,
    }, Vector)
end


function Vector.__mul(a, b)
  local new
  if getmetatable(b) == Vector then
    new = {
      x = a.x * b.x,
      y = a.y * b.y,
      z = a.z * b.z,
    }
  else
    new = {
      x = a.x * b,
      y = a.y * b,
      z = a.z * b,
    }
  end
  return setmetatable(new, Vector)
end


function Vector.__div(a, b)
  local new
  if getmetatable(b) == Vector then
    new = {
      x = a.x / b.x,
      y = a.y / b.y,
      z = a.z / b.z,
    }
  else
    new = {
      x = a.x / b,
      y = a.y / b,
      z = a.z / b,
    }
  end
  return setmetatable(new, Vector)
end


function Vector:__unm()
  return setmetatable(
    {
      x = -self.x,
      y = -self.y,
      z = -self.z,
    }, Vector)
end


function Vector.__eq(a, b)
  return a.x == b.x and a.y == b.y and a.z == b.z
end


function Vector:add(b)
  self.x = self.x + b.x
  self.y = self.y + b.y
  self.z = self.z + b.z
end


function Vector:sub(b)
  self.x = self.x - b.x
  self.y = self.y - b.y
  self.z = self.z - b.z
end


function Vector:mul(b)
  if getmetatable(b) == Vector then
    self.x = self.x * b.x
    self.y = self.y * b.y
    self.z = self.z * b.z
  else
    self.x = self.x * b
    self.y = self.y * b
    self.z = self.z * b
  end
end


function Vector:div(b)
  if getmetatable(b) == Vector then
    self.x = self.x / b.x
    self.y = self.y / b.y
    self.z = self.z / b.z
  else
    self.x = self.x / b
    self.y = self.y / b
    self.z = self.z / b
  end
end


function Vector:neg()
  self.x = -self.x
  self.y = -self.y
  self.z = -self.z
end


function Vector:normal2d()
  return setmetatable(
    {
      x = -self.y,
      y = self.x,
      z = 0,
    }, Vector)
end


function Vector:length_squared()
  return Vector.dot(self, self)
end


function Vector:length()
  return math.sqrt(Vector.dot(self, self))
end


function Vector:normalize()
  local l = self:length()
  self.x = self.x / l
  self.y = self.y / l
  self.z = self.z / l
end


function Vector:normalized()
  local l = self:length()
  return setmetatable(
    {
      x = self.x / l,
      y = self.y / l,
      z = self.z / l
    }, Vector)
end


function Vector.cross(a, b)
  return setmetatable(
    {
      x = a.y * b.z - a.z * b.y,
      y = a.z * b.x - a.x * b.z,
      z = a.x * b.y - a.y * b.x
    }, Vector)
end


function Vector.dot(a, b)
  return a.x * b.x + a.y * b.y + a.z * b.z
end


function Vector:set(x, y, z)
  if getmetatable(x) == Vector then
    self.x = x.x
    self.y = x.y
    self.z = x.z
  else
    self.x = x or 0
    self.y = y or 0
    self.z = z or 0
  end
end


function Vector:clone()
  return setmetatable(
    {
      x = self.x or 0,
      y = self.y or 0,
      z = self.z or 0
    }, Vector)
end


function Vector:rotateZ(r)
  local cr = math.cos(r)
  local sr = math.sin(r)

  local x = self.x
  local y = self.y

  self.x = x * cr - y * sr
  self.y = x * sr + y * cr
end


function Vector:rotatedZ(r)
  local v = self:clone()
  v:rotateZ(r)
  return v
end


return { Vector = newVector }
