--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

local BASE = (...):match("(.-)[^%.]+$")
entity = require(BASE .. ".entity")

local World = {}
World.__index = World


function World:registerComponentType(name, componentStore)
  if type(name) ~= "string" then
    error("Component type name must be a string")
  end

  self.componentTypes[name] = componentStore
end


function World:makeComponent(typename, ...)
  local comp = self.componentTypes[typename]:make(...)
  comp.type = typename
  return comp
end


function World:newEntity(name)
  local e = entity.Entity(self, name)
  self.toAdd[#self.toAdd+1] = e
  return e
end


function World:getEntityByName(name)
  return self.entitiesByName[name]
end


function World:foreachEntity(func)
  for k in pairs(self.entities) do
    func(k)
  end
end


local function hasRemainingComponents(ent, tab)
  for i = 2, #tab do
    if ent.components[tab[i]] == nil then
      return false
    end
  end

  return true
end


local function foreachEntityWithComponents(self, func, tab)
  -- This should speed up the search by making negative matches findable earlier.
  -- It also slows down the case with few entities or components, but that should
  -- be the exceptional case which will still be fast enough
  local comp = self.componentTypes
  table.sort(tab, function(a, b) return comp[a].count < comp[b].count end)

  for k, v in pairs(comp[tab[1]].components) do
    local ent = v.entity
    if self.entities[ent] then  -- make sure entity is actually active at the moment
      if hasRemainingComponents(ent, tab) then
        func(ent)
      end
    end
  end
end


function World:foreachEntityWithComponents(func, tab, ...)
  if type(tab) == 'table' then
    foreachEntityWithComponents(self, func, tab)
  else
    foreachEntityWithComponents(self, func, {tab, ...})
  end
end


function World:killEntity(nameOrInstance)
  if type(nameOrInstance) == "string" then
    nameOrInstance = self.entitiesByName[nameOrInstance]

    if not nameOrInstance then
      error("The world doesn't know the entity")
    end
  end

  self.toRemove[#self.toRemove+1] = nameOrInstance
end


function World:processNewEntities()
  for i = #self.toAdd, 1, -1 do
    local e = self.toAdd[i]
    if e.name then
      if self.entitiesByName[e.name] then
        error(string.format("Entity with name %s already exists in the world.", e.name))
      end

      self.entitiesByName[e.name] = e
    end

    self.entities[e] = e
  end

  self.addedLastStep = self.toAdd
  self.toAdd = {}
end


function World:processDeadEntities()
  for i = #self.toRemove, 1, -1 do
    local e = self.toRemove[i]
    self.entities[e] = nil
    if e.name then
      self.entitiesByName[e.name] = nil
    end
    e:destroy()
  end

  self.removedLastStep = self.toRemove
  self.toRemove = {}
end


function World:step()
  self:processDeadEntities()
  self:processNewEntities()
end


function World:getComponentStore(typename)
  return self.componentTypes[typename]
end


local function newWorld()
  return setmetatable({
    entities = {},
    entitiesByName = {},
    toRemove = {},
    toAdd = {},
    componentTypes = {},
    removedLastStep = {},
    addedLastStep = {}
  }, World)
end


return {World = newWorld}
