--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

package.path = "../coup/?.lua;?.lua"
local tables = require "tables"
local test = require "test"


test.group("table.merge")

:case("empty inputs", function()
  local r = tables.merge()
  assert(type(r) == "table", "result is not a table")
  assert(#r == 0, "merge of zero tables should be empty")
end)

:case("single input table", function()
  local t = {1, 2, 3, x=4, y=5}
  local s = {1, 2, 3, x=4, y=5}
  local r = tables.merge(t)
  assert(r ~= t, "function returned input table instead of new table")

  local count = 0
  for k, v in pairs(r) do
    assert(t[k] == v, "result has extra values")
    count = count +1
  end

  assert(count == 5, "number of entries in final table incorrect")

  for k, v in pairs(t) do
    assert(r[k] == v, "result is missing a value")
    assert(s[k] == v, "input table was changed")
  end

  for k, v in pairs(s) do
    assert(t[k] == v, "Input table was changed")
  end
end)

:case("overwrite several times", function()
  local r = tables.merge({x = 1, a = 1},
                         {x = 2, b = 1},
                         {x = 3, c = 1},
                         {x = 4, d = 1},
                         {x = 5, e = 1})

  assert(r.x == 5, "overwritten value not correct")
  assert(r.a == 1, "copied value not correct")
  assert(r.b == 1, "copied value not correct")
  assert(r.c == 1, "copied value not correct")
  assert(r.d == 1, "copied value not correct")
  assert(r.e == 1, "copied value not correct")
end)


test.group("table.scoped_set/get")

:case("get/set", function()
  local t = {x = 1,
             y = {
               z = { }
             }
           }

  assert(tables.scoped_set(t, "a", 10) == true,           "scoped_set did not return true")
  assert(t.a == 10,                                       "Value in top level not set correctly")
  assert(tables.scoped_get(t, "a") == 10,                 "Value in top level not returned correctly")
  assert(tables.scoped_get(t, "b") == nil,                "Missing value did not return nil")
  assert(tables.scoped_set(t, "y.z.b", 20) == true,       "scoped_set did not return true")
  assert(t.y.z.b == 20,                                   "Value in nested table not set correctly")
  assert(tables.scoped_get(t, "y.z.b") == 20,             "Value in nested table not returned correctly")
  assert(tables.scoped_set(t, "y.g.b", 10) == false,      "scoped_set did not return false")
  assert(tables.scoped_set(t, "y.g.b", 10, true) == true, "scoped_set did not return true")
  assert(t.y.g.b == 10,                                   "scoped_set did not correctly generate tables")
end)


test.group("Mapping functions")

:case("map_iter with ipairs", function()
  local values = {}
  local input = {1,2,3,4,5,6}
  local double = function(x) return 2*x end
  for k, v in tables.map_iter(double, ipairs(input)) do
    values[k] = v
  end

  assert(#values == #input)
  for i = 1,6 do
    assert(values[i] == 2*input[i])
  end
end)


:case("map_iter with pairs", function()
  local values = {}
  local input = {a = 1, b = 2, c = 3, d = 4, e = 5, f = 6}
  local double = function(x) return 2*x end
  for k, v in tables.map_iter(double, pairs(input)) do
    values[k] = v
  end

  assert(values.a ==  2)
  assert(values.b ==  4)
  assert(values.c ==  6)
  assert(values.d ==  8)
  assert(values.e == 10)
  assert(values.f == 12)
end)


:case("map_iter chaining", function()
  local values = {}
  local input = {1,2,3,4,5,6}
  local double = function(x) return 2*x end
  for k, v in tables.map_iter(double, tables.map_iter(double, ipairs(input))) do
    values[k] = v
  end

  assert(#values == #input)
  for i = 1,6 do
    assert(values[i] == 4*input[i])
  end
end)


:case("map_array", function()
  local values = {}
  local input = {1,2,3,4,5,6}
  local double = function(x) return 2*x end
  values = tables.map_array(double, input)

  assert(#values == #input)
  for i = 1,6 do
    assert(values[i] == 2*input[i])
  end

  local res2 = tables.map_array(double, {})
  assert(#res2 == 0)
end)


:case("map", function()
  local values = {}
  local input = {a=1,b=2,c=3,d=4,e=5,f=6}
  local double = function(x) return 2*x end
  values = tables.map(double, input)

  local c = 0
  for k, v in pairs(values) do
    c = c + 1
  end
  assert(c == 6)

  for k in pairs(input) do
    assert(values[k] == 2*input[k])
  end

  local res2 = tables.map(double, {})
  c = 0
  for k, v in pairs(res2) do
    c = c + 1
  end
  assert(c == 0)

end)


:case("collect", function()
  local input = {a = 1, b = 2, c = 3, 6, 7, 8}
  local output = tables.collect(pairs(input))

  assert(output.a == 1)
  assert(output.b == 2)
  assert(output.c == 3)
  assert(output[1] == 6)
  assert(output[2] == 7)
  assert(output[3] == 8)

  local cnt = 0
  for k, v in pairs(output) do
    cnt = cnt + 1
  end
  assert(cnt == 6)
end)


test.group("Reduce")

:case("reduce_iter", function()
  local sum = function(a,b) return a + b end
  local input = {1,2,3,4,5,6}
  assert(tables.reduce_iter(sum, 0,   ipairs(input)) == 21)
  assert(tables.reduce_iter(sum, nil, ipairs(input)) == 21)
  assert(tables.reduce_iter(sum, 0,   ipairs({}))    ==  0)
  test.assert_error(function() tables.reduce_iter(sum, nil, ipairs({})) end)
end)


:case("reduce", function()
  local sum = function(a,b) return a + b end
  local input = {1,2,3,4,5,6}
  assert(tables.reduce(sum, input, 0) == 21)
  assert(tables.reduce(sum, input)    == 21)
  assert(tables.reduce(sum, {},    0) ==  0)
  test.assert_error(function() tables.reduce(sum, {}) end)
end)


:case("filter_iter with ipairs", function()
  local input = {1,2,3,4,5,6,7,8,9,10}
  local output = {}
  function odd(v) return v % 2 == 1 end
  for k, v in tables.filter_iter(odd, ipairs(input)) do
    output[#output+1] = v
  end
  assert(#output == 5)
  assert(output[1] == 1)
  assert(output[2] == 3)
  assert(output[3] == 5)
  assert(output[4] == 7)
  assert(output[5] == 9)
end)


:case("filter_iter with pairs", function()
  local input = {1,2,3,4,5,6,7,8,9,10}
  local output = {}
  function odd(v) return v % 2 == 1 end
  for k, v in tables.filter_iter(odd, pairs(input)) do
    output[#output+1] = v
  end
  assert(#output == 5)
  assert(output[1] == 1)
  assert(output[2] == 3)
  assert(output[3] == 5)
  assert(output[4] == 7)
  assert(output[5] == 9)
end)


:case("filter_array", function()
  local input = {1,2,3,4,5,6,7,8,9,10}
  function odd(v) return v % 2 == 1 end
  local output =  tables.filter_array(odd, input)

  assert(#output == 5)
  assert(output[1] == 1)
  assert(output[2] == 3)
  assert(output[3] == 5)
  assert(output[4] == 7)
  assert(output[5] == 9)
end)


:case("filter", function()
  local input = {1,2,3,4,5,6,7,8,9,10}
  function odd(v) return v % 2 == 1 end
  local output =  tables.filter(odd, input)

  local cnt = 0
  for k, v in pairs(output) do
    cnt = cnt + 1
  end
  assert(cnt == 5)

  assert(output[1] == 1)
  assert(output[3] == 3)
  assert(output[5] == 5)
  assert(output[7] == 7)
  assert(output[9] == 9)
end)


test.group("tables.concat")

:case("Empty input", function()
  local res = tables.concat()
  assert(#res == 0)
end)


:case("Empty input arrays", function()
  local res = tables.concat({}, {}, {})
  assert(#res == 0)
end)


:case("Normal use", function()
  local res = tables.concat({1, 2, 3}, {4, 5, 6}, {7, 8, 9})
  assert(#res == 9)
  for i = 1, 9 do
    assert(res[i] == i)
  end
end)


:case("Unchanged inputs?", function()
  local i1 = {1, 2, 3}
  local i2 = {4, 5, 6}

  local res = tables.concat(i1, i2)
  assert(res ~= i1)
  assert(res ~= i2)

  for i = 1, 3 do
    assert(i1[i] == i)
    assert(i2[i] == i+3)
  end
end)


test.group("tables.append")

:case("Empty input", function()
  local input = {1, 2, 3}
  tables.append(input)

  assert(#input == 3)
  for i = 1, 3 do
    assert(input[i] == i)
  end
end)


:case("Normal use", function()
  local input = {1, 2, 3}
  tables.append(input, 4, 5, 6)
  assert(#input == 6)
  for i = 1, 6 do
    assert(input[i] == i)
  end
end)


test.group("tables.append_all")

:case("Empty input", function()
  local input = {1, 2, 3}
  tables.append_all(input)

  assert(#input == 3)
  for i = 1, 3 do
    assert(input[i] == i)
  end
end)


:case("Normal use", function()
  local input = {1, 2, 3}
  tables.append_all(input, {4, 5, 6}, {7, 8, 9})
  assert(#input == 9)
  for i = 1, 9 do
    assert(input[i] == i)
  end
end)

:case("first_iter with ipairs", function()
  local input = {1,1,2,3,4,5,6,7,8,9,10}
  local gt4 = function(x) return x > 4 end
  local k, v = tables.first_iter(gt4, ipairs(input))
  assert(k == 6)
  assert(v == 5)

  k, v = tables.first_iter(gt4, ipairs({}))
  assert(k == nil)
  assert(v == nil)

  k, v = tables.first_iter(function() return false end, ipairs(input))
  assert(k == nil)
  assert(v == nil)
end)


:case("first", function()
  local input = {1,1,2,3,4,5,6,7,8,9,10}
  local gt4 = function(x) return x > 4 end
  local k, v = tables.first(gt4, input)
  assert(k == 6)
  assert(v == 5)

  k, v = tables.first(gt4, {})
  assert(k == nil)
  assert(v == nil)

  k, v = tables.first(function() return false end, input)
  assert(k == nil)
  assert(v == nil)
end)


:case("keys", function()
  assert("false") -- need to implement the test
end)


:case("random_table", function()
  assert("false") -- need to implement the test
end)


:case("random_array", function()
  assert("false") -- need to implement the test
end)


:case("Array2D", function()
  assert("false") -- need to implement the test
end)


test.group("deep_equal")

:case("booleans", function()
  assert(tables.deep_equal(true, true))
  assert(not tables.deep_equal(true, false))
end)

:case("numbers", function()
  assert(tables.deep_equal(1, 1))
  assert(not tables.deep_equal(1, 2))
end)

:case("strings", function()
  assert(tables.deep_equal("foo", "foo"))
  assert(not tables.deep_equal("foo", "bar"))
  assert(not tables.deep_equal("10", 10))
end)

:case("functions", function()
  local function f_a() end
  local function f_b() end
  assert(tables.deep_equal(f_a, f_a))
  assert(not tables.deep_equal(f_a, f_b))
end)

:case("userdata", function()
  local c_a = love.graphics.newCanvas()
  local c_b = love.graphics.newCanvas()
  assert(tables.deep_equal(c_a, c_a))
  assert(not tables.deep_equal(c_a, c_b))
end)

:case("threads", function()
  local function f_a() end
  local function f_b() end
  local t_a = coroutine.create(f_a)
  local t_b = coroutine.create(f_b)
  assert(tables.deep_equal(t_a, t_a))
  assert(not tables.deep_equal(t_a, t_b))
end)

:case("flat arrays", function()
  assert(tables.deep_equal({1,2,3,4}, {1,2,3,4}))
  assert(not tables.deep_equal({1,2,3,4,5}, {1,2,3,4}))
  assert(not tables.deep_equal({1,2,3,4}, {1,2,3,4,5}))
end)

:case("flat tables", function()
  assert(tables.deep_equal({a = 1, b = 2, c = 3}, {a = 1, b = 2, c = 3}))
  assert(not tables.deep_equal({a = 1, b = 2, c = 4}, {a = 1, b = 2, c = 3}))
  assert(not tables.deep_equal({a = 1, b = 2}, {a = 1, b = 2, c = 3}))
  assert(not tables.deep_equal({a = 1, b = 2, c = 3}, {a = 1, b = 2}))
end)

:case("nested tables", function()
  assert(tables.deep_equal({{{}}}, {{{}}}))
  assert(tables.deep_equal({a={b={c=1, d=true, e="foo"}}}, {a={b={c=1, d=true, e="foo"}}}))
  assert(not tables.deep_equal({}, {{}}))
end)

:case("self-referencing tables", function()
  local t = {x = 10}
  t[1] = t
  local t2 = {x = 10}
  t2[1] = t2
  assert(tables.deep_equal(t, t2))
  t2[2] = 20
  assert(not tables.deep_equal(t, t2))
end)

:case("complex keys", function()
  local function f() end
  assert(tables.deep_equal({[f] = 1}, {[f] = 1}))

  --[[
    local t_a = { 1 }
    local t_b = { 1 }
    assert(tables.deep_equal({[t_a] = 1}, {[t_b] = 1}))

    local t_c = { 2 }
    assert(not tables.deep_equal({[t_a] = 1}, {[t_c] = 1}))
  --]]
end)


test.group("tables.array_find")

:case("value exists", function()
  assert(tables.array_find({5,4,3,2,1,0}, 1) == 5)
end)


:case("value does not exist", function()
  assert(tables.array_find({5,4,3,2,1,0}, 6) == nil)
end)


test.group("tables.array_find_f")

:case("value exists", function()
  assert(tables.array_find_f({5,4,3,2,1,0}, function(v) return v == 1 end) == 5)
end)


:case("value does not exist", function()
  assert(tables.array_find({5,4,3,2,1,0}, function(v) return false end) == nil)
end)
