--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

package.path = "../coup/?.lua;?.lua"
local tween = require "tween"
local test  = require "test"

test.group("tween")

:case("simple tween", function()
  local group = tween.group()
  local x = {x = 0}
  group:tween(x, 1, {x = 1})
  assert(x.x == 0)
  group:update(0.1)
  assert(x.x == 0.1)
  group:update(0.4)
  assert(x.x == 0.5)
  group:update(0.3)
  assert(x.x == 0.8)
  group:update(0.2)
  assert(x.x == 1.0)
  group:update(1.0)
  assert(x.x == 1.0)
end)


:case("multiple tweens", function()
  local group = tween.group()
  local x = {x = 0}
  local y = {y = 1, z = 2}

  group:tween(x, 1, {x = 2})
  group:tween(y, 2, {y = 2})
  group:tween(y, 4, {z = 0})

  assert(x.x == 0)
  assert(y.y == 1)
  assert(y.z == 2)

  group:update(0.1)

  assert(x.x == 0.2)
  assert(y.y == 1.05)
  assert(y.z == 1.95)

  group:update(0.9)

  assert(x.x == 2)
  assert(y.y == 1.5)
  assert(y.z == 1.5)

  group:update(1.0)

  assert(x.x == 2)
  assert(y.y == 2)
  assert(y.z == 1)

  group:update(1.0)

  assert(x.x == 2)
  assert(y.y == 2)
  assert(y.z == 0.5)

  group:update(0.5)

  assert(x.x == 2)
  assert(y.y == 2)
  assert(y.z == 0.25)

  group:update(0.5)

  assert(x.x == 2)
  assert(y.y == 2)
  assert(y.z == 0)

  group:update(1.0)

  assert(x.x == 2)
  assert(y.y == 2)
  assert(y.z == 0)
end)


:case("multiple vars in one tween", function()
  local group = tween.group()
  local y = {y = 0, z = 0}

  group:tween(y, 1, {y = 10, z = 100})

  assert(y.y == 0)
  assert(y.z == 0)

  group:update(0.5)

  assert(y.y == 5)
  assert(y.z == 50)

  group:update(0.25)

  assert(y.y == 7.5)
  assert(y.z == 75)

  group:update(0.3)

  assert(y.y == 10)
  assert(y.z == 100)

  group:update(1)

  assert(y.y == 10)
  assert(y.z == 100)
end)


:case("chaining tweens with :after", function()
  local group = tween.group()
  local x = {x = 0}

  group:tween(x, 1, {x = 5})
    :after(1, {x = 2})
    :after(2, {x = 10})

  assert(x.x == 0)
  group:update(0.5)
  assert(x.x == 2.5)
  group:update(1)
  assert(x.x == 3.5)
  group:update(1.5)
  assert(x.x == 6)
  group:update(1)
  assert(x.x == 10)
  group:update(1)
  assert(x.x == 10)
end)


:case("chaining tweens for different objects with :after", function()
  local group = tween.group()
  local x = {x = 0}
  local y = {y = 0}

  group:tween(x, 1, {x = 1})
    :after(y, 1, {y = 1})

  assert(x.x == 0)
  assert(y.y == 0)

  group:update(0.5)
  assert(x.x == 0.5)
  assert(y.y == 0)

  group:update(0.25)
  assert(x.x == 0.75)
  assert(y.y == 0)

  group:update(1.0)
  assert(x.x == 1.0)
  assert(y.y == 0.75)

  group:update(1.0)
  assert(x.x == 1.0)
  assert(y.y == 1.0)

  group:update(1.0)
  assert(x.x == 1.0)
  assert(y.y == 1.0)
end)


:case(":delay'ing tweens", function()
  local group = tween.group()
  local x = {x = 0}

  group:tween(x, 1, {x = 1})
    :delay(1)

  assert(x.x == 0)
  
  group:update(0.5)
  assert(x.x == 0)
  group:update(0.25)
  assert(x.x == 0)
  group:update(0.5)
  assert(x.x == 0.25)
end)


:case(":on_start, :on_update and :on_complete", function()
  local group = tween.group()
  local x = {x = 0}

  local update = 0
  local start = 0
  local complete = 0

  group:tween(x, 1, {x = 1})
    :delay(1)
    :on_start(function() start = start + 1 end)
    :on_update(function() update = update + 1 end)
    :on_complete(function() complete = complete + 1 end)

  assert(update == 0)
  assert(start == 0)
  assert(complete == 0)
  
  group:update(0.5)
  assert(update == 0)
  assert(start == 0)
  assert(complete == 0)

  group:update(0.75)
  assert(update == 1)
  assert(start == 1)
  assert(complete == 0)

  group:update(0.75)
  assert(update == 2)
  assert(start == 1)
  assert(complete == 1)
end)


:case(":abort", function()
  local group = tween.group()
  local x = {x = 0}
  local y = {y = 0}

  local update = 0
  local complete = 0

  local tween = group:tween(x, 1, {x = 1})
    :on_update(function() update = update + 1 end)
    :on_complete(function() complete = complete + 1 end)
  group:tween(y, 1, {y = 1})

  assert(update == 0)
  assert(complete == 0)

  group:update(0.2)
  assert(update == 1)
  assert(complete == 0)

  tween:abort()
  assert(update == 1)
  assert(complete == 0)
  assert(y.y == 0.2)
  assert(x.x == 0.2)

  group:update(0.2)
  assert(y.y == 0.4)
  assert(x.x == 0.2)

  group:update(10)
  assert(update == 1)
  assert(complete == 0)
  assert(y.y == 1.0)
  assert(x.x == 0.2)
end)
