--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

package.path = "../coup/?.lua;?.lua"
local vector = require "vector"
local test = require "test"

test.group("vector math")

:case("constructor", function()
  local v = vector.Vector()
  assert(v.x == 0)
  assert(v.y == 0)
  assert(v.z == 0)

  v = vector.Vector(2)
  assert(v.x == 2)
  assert(v.y == 0)
  assert(v.z == 0)

  v = vector.Vector(2, 3)
  assert(v.x == 2)
  assert(v.y == 3)
  assert(v.z == 0)

  v = vector.Vector(2, 3, 4)
  assert(v.x == 2)
  assert(v.y == 3)
  assert(v.z == 4)
end)


:case("add", function()
  local v = vector.Vector(1, 2, 3) + vector.Vector(4, 5, 6)
  assert(v.x == 5)
  assert(v.y == 7)
  assert(v.z == 9)
end)


:case("assignment add", function()
  local v = vector.Vector(1, 2, 3)
  v:add(vector.Vector(4, 5, 6))
  assert(v.x == 5)
  assert(v.y == 7)
  assert(v.z == 9)
end)


:case("subtract", function()
  local v = vector.Vector(1, 2, 3) - vector.Vector(4, 3, 1)
  assert(v.x == -3)
  assert(v.y == -1)
  assert(v.z ==  2)
end)


:case("assignment subtract", function()
  local v = vector.Vector(1, 2, 3)
  v:sub(vector.Vector(4, 3, 1))
  assert(v.x == -3)
  assert(v.y == -1)
  assert(v.z ==  2)
end)


:case("multiply - component wise", function()
  local v = vector.Vector(1, 2, 3) * vector.Vector(4, 5, 6)
  assert(v.x ==  4)
  assert(v.y == 10)
  assert(v.z == 18)
end)


:case("assignment multiply - component wise", function()
  local v = vector.Vector(1, 2, 3)
  v:mul(vector.Vector(4, 5, 6))
  assert(v.x ==  4)
  assert(v.y == 10)
  assert(v.z == 18)
end)


:case("multiply - scalar", function()
  local v = vector.Vector(1, 2, 3) * 5
  assert(v.x ==  5)
  assert(v.y == 10)
  assert(v.z == 15)
end)


:case("assignment multiply - scalar", function()
  local v = vector.Vector(1, 2, 3)
  v:mul(5)
  assert(v.x ==  5)
  assert(v.y == 10)
  assert(v.z == 15)
end)


:case("divide - component wise", function()
  local v = vector.Vector(12, 20, 30) / vector.Vector(4, 5, 6)
  assert(v.x == 3)
  assert(v.y == 4)
  assert(v.z == 5)
end)


:case("assignment divide - component wise", function()
  local v = vector.Vector(12, 20, 30)
  v:div(vector.Vector(4, 5, 6))
  assert(v.x == 3)
  assert(v.y == 4)
  assert(v.z == 5)
end)


:case("divide - scalar", function()
  local v = vector.Vector(10, 15, 20) / 5
  assert(v.x == 2)
  assert(v.y == 3)
  assert(v.z == 4)
end)


:case("assignment divide - scalar", function()
  local v = vector.Vector(10, 15, 20)
  v:div(5)
  assert(v.x == 2)
  assert(v.y == 3)
  assert(v.z == 4)
end)


:case("unary minus", function()
  local v = -vector.Vector(10, 15, 20)
  assert(v.x == -10)
  assert(v.y == -15)
  assert(v.z == -20)
end)


:case("assignment unary minus ('negate')", function()
  local v = vector.Vector(10, 15, 20)
  v:neg()
  assert(v.x == -10)
  assert(v.y == -15)
  assert(v.z == -20)
end)


:case("equality", function()
  assert(vector.Vector(1,2,3) == vector.Vector(1,2,3))
end)


:case("dot product", function()
  assert(vector.Vector(1,2,3):dot(vector.Vector(2,3,4)) == 20)
end)


:case("normal", function()
  local v = vector.Vector(3,7)
  local n = v:normal2d()
  assert(math.abs(v:dot(n)) < 0.0000001)
end)


:case("length", function()
  local l = vector.Vector(2,3,4)
  assert(math.abs(l:length() - math.sqrt(29)) <= 0.0000001)
end)


:case("length_squared", function()
  local l = vector.Vector(2,3,4)
  assert(l:length_squared() == 29)
end)


:case("normalized", function()
  assert(math.abs(vector.Vector(1,2,3):normalized():length_squared() - 1.0) < 0.00000001)
end)


:case("normalize", function()
  local v = vector.Vector(1,2,3)
  v:normalize()
  assert(math.abs(v:length_squared() - 1.0) < 0.00000001)
end)


:case("cross", function()
  local v = vector.Vector(1,0,0):cross(vector.Vector(0,1,0))
  assert(v == vector.Vector(0,0,1))

  local v = vector.Vector(1,2,3):cross(vector.Vector(3,5,6))
  assert(v == vector.Vector(-3, 3, -1))
end)


:case("set from vector", function()
  local v = vector.Vector()
  v:set(vector.Vector(1,2,3))
  assert(v == vector.Vector(1,2,3))
end)


:case("set from arguments", function()
  local v = vector.Vector()
  v:set(1, 2, 3)
  assert(v == vector.Vector(1,2,3))

  v:set(1)
  assert(v == vector.Vector(1, 0, 0))

  v:set(1, 1)
  assert(v == vector.Vector(1, 1, 0))
end)


:case("clone", function()
  local v = vector.Vector(1,2,3)
  local w = v:clone()
  
  assert(v == w)
  w.z = 0

  -- make sure an actual new object was returned
  assert(v ~= w)
end)


:case("rotateZ", function()
  local v = vector.Vector(1,2,3)
  local w = v:clone()
  w:rotateZ(0)
  assert(v == w)

  w:rotateZ(0.5 * math.pi)
  assert((w - vector.Vector(-2, 1, 3)):length() < 0.00001)

  assert(v:rotatedZ(0.5 * math.pi) == w)
  assert(v == vector.Vector(1,2,3))
end)
