--
-- coup
--
-- Copyright (c) 2014-2016, rnlf
--
-- This library is free software; you can redistribute it and/or modify it
-- under the terms of the MIT license. See LICENSE for details.
--

package.path = "../coup/?.lua;?.lua"
local world = require "world"
local ComponentStore = require "components.componentstore"
local test = require "test"


test.group("World")

:case("basic entity functionality", function()
  local w = world.World()

  local e = w:newEntity("Foo")

  local countEntities = function()
    local c = 0
    w:foreachEntity(function()
      c = c + 1
    end)
    return c
  end

  assert(#w.addedLastStep == 0)

  assert(countEntities() == 0)
  assert(w:getEntityByName("Foo") == nil)
  w:step()
  assert(#w.addedLastStep == 1)
  assert(w.addedLastStep[1] == e)
  assert(countEntities() == 1)
  assert(w:getEntityByName("Foo") == e)
  w:step()
  assert(#w.addedLastStep == 0)
  assert(countEntities() == 1)
  w:killEntity(e)
  assert(#w.removedLastStep == 0)
  assert(w:getEntityByName("Foo") == e)
  assert(countEntities() == 1)
  w:step()
  assert(#w.removedLastStep == 1)
  assert(w.removedLastStep[1] == e)
  assert(countEntities() == 0)
  assert(w:getEntityByName("Foo") == nil)
end)


:case("call destroy on kill", function()
  local destroyed

  local w = world.World()
  local e = w:newEntity()

  e.destroy = function(self)
    destroyed = self
  end

  w:killEntity(e)
  assert(destroyed == nil)
  w:step()

  assert(destroyed == e)

end)


:case("component management", function()
  local w = world.World()

  local function make(x, y)
    return {x = x, y = y}
  end

  local store = ComponentStore(make)
  w:registerComponentType("Test", store)

  assert(store == w:getComponentStore("Test"))


  local e = w:newEntity("Foo")
  local c = e:makeComponent("Test", 10, 20)

  assert(c.x == 10)
  assert(c.y == 20)

  assert(e.components["Test"] == c)
end)


:case("component management errors", function()
  local w = world.World()

  local function make() end
  test.assert_error(function() w:registerComponentType(100, ComponentStore(make)) end)
end)


:case("Component search", function()
  local w = world.World()

  local function make()
    return {}
  end

  w:registerComponentType("Test", ComponentStore(make))
  w:registerComponentType("Test2", ComponentStore(make))
  w:registerComponentType("Test3", ComponentStore(make))

  local e1 = w:newEntity("A")
  e1:makeComponent("Test")
  local e2 = w:newEntity("B")
  e2:makeComponent("Test")
  e2:makeComponent("Test2")
  local e3 = w:newEntity("C")
  e3:makeComponent("Test")
  e3:makeComponent("Test2")
  e3:makeComponent("Test3")

  local found = {}
  local foundCount = 0

  local function process(e)
    found[e] = e
    foundCount = foundCount + 1
  end

  -- Before stepping, nothing should be found

  w:foreachEntityWithComponents(process, {"Test"})
  assert(foundCount == 0)

  -- without constraints
  found = {}
  foundCount = 0

  w:foreachEntity(process)
  assert(foundCount == 0)

  w:step()

  w:foreachEntityWithComponents(process, {"Test"})
  assert(foundCount == 3)
  assert(found[e1] == e1)
  assert(found[e2] == e2)
  assert(found[e3] == e3)

  found = {}
  foundCount = 0

  w:foreachEntityWithComponents(process, {"Test2", "Test3"})
  assert(foundCount == 1)
  assert(found[e3] == e3)

  found = {}
  foundCount = 0

  w:foreachEntityWithComponents(process, {"Test", "Test2"})
  assert(foundCount == 2)
  assert(found[e2] == e2)
  assert(found[e3] == e3)


  -- without table
  found = {}
  foundCount = 0

  w:foreachEntityWithComponents(process, "Test")
  assert(foundCount == 3)
  assert(found[e1] == e1)
  assert(found[e2] == e2)
  assert(found[e3] == e3)

  found = {}
  foundCount = 0

  w:foreachEntityWithComponents(process, "Test2", "Test3")
  assert(foundCount == 1)
  assert(found[e3] == e3)

  found = {}
  foundCount = 0

  w:foreachEntityWithComponents(process, "Test", "Test2")
  assert(foundCount == 2)
  assert(found[e2] == e2)
  assert(found[e3] == e3)

  -- without constraints
  found = {}
  foundCount = 0

  w:foreachEntity(process)
  assert(foundCount == 3)
  assert(found[e1] == e1)
  assert(found[e2] == e2)
  assert(found[e3] == e3)

  -- remove entity
  w:killEntity(e2)

  -- should find the same until world:step()
  found = {}
  foundCount = 0

  w:foreachEntity(process)
  assert(foundCount == 3)
  assert(found[e1] == e1)
  assert(found[e2] == e2)
  assert(found[e3] == e3)

  w:step()

  found = {}
  foundCount = 0

  w:foreachEntity(process)
  assert(foundCount == 2)
  assert(found[e1] == e1)
  assert(found[e3] == e3)

  -- same for constrained version
  found = {}
  foundCount = 0

  w:foreachEntityWithComponents(process, "Test")
  assert(foundCount == 2)
  assert(found[e1] == e1)
  assert(found[e3] == e3)

end)
