local debugging = require "debugging"


function createMonster(options)
  options = options or {}
  bobAmplitude = options.bobAmplitude or 10
  bobFrequency = options.bobFrequency or 5

  local pos = {-1000, -1000}
  local monster = {
    pos = pos,
    speed = options.speed or 100,
    paused = options.paused or false,
    hidden = options.hidden or false,
    avoidance = options.avoidance or 300,
    paths = options.paths or {},
    _waypoints = {pos},
    _prev = pos,
    _next = pos,
    _index = 1,
    _travelled = -math.huge,
    _distance = 1,
    _totalTime = 0,
  }

  function distance(a, b)
    return math.sqrt(coup.smath.sq(a[1] - b[1]) + coup.smath.sq(a[2] - b[2]))
  end

  function monster:angleTowards(vector, pos)
    local pos = pos or self.pos
    local dx = vector.x - pos[1]
    return math.atan(((vector.y - 40) - pos[2]) / dx)
  end

  function monster:attainWaypoint(index, travelled)
    self._index = index
    self._prev = self._waypoints[index]
    self._next = self._waypoints[index + 1]
    self._travelled = travelled or 0
    if self._next == nil then
      -- We've reached the end of the polyline.
      -- Set _next to _prev so we stay still.
      self._next = self._prev
      -- Set distance to -infinity so we never reach the next point.
      self._distance = math.huge
    else
      self._distance = distance(self._prev, self._next)
    end
  end

  function monster:setWaypoints(waypoints)
    self._waypoints = waypoints
    self:attainWaypoint(1, 0)
    self:reposition()
  end

  function monster:loadWaypointsFromPolylineObject(obj, x)
    assert(obj, 'Invalid object')
    assert(obj.shape == 'polyline', 'Invalid object for waypoints')
    local waypoints = {}
    for k, point in ipairs(obj.polyline) do
      table.insert(waypoints, {point.x + obj.x, point.y + obj.y})
    end
    self:setWaypoints(waypoints)
  end

  function monster:move(value)
    self.paused = not value
    print('monster paused', not value)
  end

  function monster:show(value)
    self.hidden = not value
  end

  function monster:handleEvent(event)
    if event.path then
      local pathname = event.path
      local path = nil
      for i, polyline in ipairs(self.paths) do
        if polyline.name == pathname then
          path = polyline
          break
        end
      end
      self:loadWaypointsFromPolylineObject(path)
    end
    if event.moving ~= nil then
      -- print('move', event.moving)
      self:move(event.moving)
    end
    if event.hidden ~= nil then
      -- print('show', not event.hidden)
      self:show(not event.hidden)
    end
    if event.avoidance then
      -- print('avoidance', event.avoidance)
      self.avoidance = event.avoidance
    end
    if event.speed then
      -- print('speed', event.speed)
      self.speed = event.speed
    end
  end

  function monster:update(dt)
    if self.paused then
      return
    end
    self._travelled = self._travelled + dt * self.speed
    if self._travelled > self._distance then
      self:attainWaypoint(self._index + 1, self._travelled - self._distance)
    end
    self._totalTime = self._totalTime + dt
    self:reposition()
  end

  function monster:reposition()
    local bob = bobAmplitude * math.sin(bobFrequency * self._totalTime)
    local ratio = self._travelled / self._distance
    self.pos[1] = coup.smath.lerp(self._prev[1], self._next[1], ratio)
    self.pos[2] = coup.smath.lerp(self._prev[2], self._next[2], ratio) + bob
  end

  return monster
end


return createMonster
