local loader = {}
local DEG_TO_RAD = math.pi / 180


-- Reorient a tilemap object.
--
-- Returns an object with:
-- - x, y coordinates equal to the object's centre if a rectangle or circle; if
--   a polygon, uses the original coordinates.
-- - Adds the angle in radians.
function loader:reorient(obj)
  local origX = obj.x
  local origY = obj.y
  local dx = 0.5 * obj.width
  local dy = 0.5 * obj.height
  if obj.gid ~= nil then
    dy = -dy
  end
  local radians = -DEG_TO_RAD * (obj.rotation or 0)
  local cos = math.cos(radians)
  local sin = math.sin(radians)
  local x = origX + dx * cos + dy * sin
  local y = origY - dx * sin + dy * cos

  return {
    type = obj.type,
    x = x,
    y = y,
    angle = radians,
    name = obj.name,
    gid = obj.gid,
    shape = obj.shape,
    properties = obj.properties,
    polygon = obj.polygon,
    width = obj.width,
    height = obj.height,
    orig = obj,
  }
end


function loader:addRectangle(hc, obj)
  local shape = hc:addRectangle(
    obj.x - obj.width / 2,
    obj.y - obj.height / 2,
    obj.width,
    obj.height
  )
  shape:rotate(-obj.angle)
  return shape
end


function loader:addPolygon(hc, obj)
  local args = {}
  for k, pos in ipairs(obj.polygon) do
    table.insert(args, obj.x + pos.x)
    table.insert(args, obj.y + pos.y)
  end

  local shape = hc:addPolygon(unpack(args))
  -- shape:rotate(-obj.angle, 0, 0)
  return shape
end


function loader:addCircle(hc, obj)
  local w = obj.width
  assert(w == obj.height, "True ellipses aren't allowed")

  -- DON'T ROTATE CIRCLES IN THE TILEMAP >:(
  return hc:addCircle(obj.x, obj.y, w / 2)
end


loader.shapeAdders = {
  rectangle = loader.addRectangle,
  ellipse = loader.addCircle,
  polygon = loader.addPolygon,
}


function loader:addShape(hc, obj)
  return self.shapeAdders[obj.shape](self, hc, obj)
end


function loader:addTerrainFromObjectLayer(hc, layer, group)
  local shapes = {}
  local reorientedObjects = {}
  for i, obj in ipairs(layer.objects) do
    local reoriented = self:reorient(obj)
    table.insert(reorientedObjects, reoriented)
    local shape = self.shapeAdders[reoriented.shape](self, hc, reoriented)
    shape.islevel = true -- See physics_updater.lua
    table.insert(shapes, shape)
    hc.addToGroup(group, shape)
    hc.setPassive(shape)
  end
  return result
end


return loader
