local Level = coup.class({})
require "events"
require "physics_updater"
require "tilemap_collider"
require "player_controller"
require "player_processor"
HighlightRenderer = require "highlight_renderer"
SpriteRenderer = require "spriterenderer"
local Sparks = require "sparks"
local Dust = require "dust"
local polygonLoader = require "tilemap_polygon_loader"
local debugging = require "debugging"
local createMonster = require "monster"
-- local buildCollisionMesh = require "tilemap_collision_builder"

local worldFilename = "data/world.lua"

function loadTiles(name)
  local filename = "data/" .. name
  return G.loadImage(filename)
end


local torch_range = 400
local fear_timeout = 5
local fear_decay_rate = 20
local eye_range_offset = 80
local min_flicker_fear = 200
local eye_strike_distance = 120
local min_drop_fear = 0.7
local max_drop_fear = 1.5

function fear_from_distance(dist)
  return (torch_range + eye_range_offset) - dist
end

local eye_strike_fear = fear_from_distance(eye_strike_distance)


function createTrigger(level, obj)
  local trigger = polygonLoader:reorient(obj)
  trigger.item = obj.type
  local shape = polygonLoader:addShape(level.collider, trigger)
  trigger.shape = shape
  shape.owner = trigger
  shape.isSensor = true
end

local objectCreators = {
  trigger = createTrigger,
  leveltrigger = createTrigger,
  falltrigger = createTrigger,
}

function objectCreators.player(level, obj)
  local e = level.world:newEntity("player")
  e:makeComponent("Transform",   obj.x, obj.y )
  e:makeComponent("Sprite",      G.animRepo:newSprite("Idle"), 2, 64, 128)
  e:makeComponent("Velocity", 0, 0)
  --e:makeComponent("PhysicsObject", level.collider, -1, -10, 11.5, 19)
  local po = e:makeComponent("PhysicsObject", level.collider, 0, -64, 
    {
       -48/2, -128,
        48/2, -128,
        48/2,  0,
       -48/2,  0
    }
  --[[
    {
      -9/2, -19/2,
      9/2, -19/2,
      11.5/2, 19/2,
      -11.5/2, 19/2
    }]]
  )
  po.item = "player"
  e:makeComponent("Input")
  local sensors = e:makeComponent("Sensors")
  sensors:addSensor(level.collider, "foot", 0, 0, 44, 4)
  sensors:addSensor(level.collider, "left", -7, -19/2, 3, 17)
  sensors:addSensor(level.collider, "right", 7, -19/2, 3, 17)
  level.activeCP.pos = {obj.x, obj.y}
  level.player = e
  e.fear = 0
  e.fear_timeout = 0
  e.torch_on = true
  e.last_contact = 0
  e.last_heartbeat = 0
end


function Level:createObjects(objs)
  for i, obj in ipairs(objs.objects.objects) do
    objectCreators[obj.type](self, obj)
  end
  for i, obj in ipairs(objs.triggers.objects) do
    objectCreators[obj.type](self, obj)
  end
end


function Level:new()
  self.totalTime = 0
  local objs, imgs, props
  self.t = 0

  self.tilemap, objs, imgs, props = coup.tiledloader.loadTiledMapFile(worldFilename, loadTiles)

  self.collider = HC.new(100)
  self.activeCP = {}

  -- XXX: render layers
  -- self.mainLayer = self.tilemap:get_layer("main")

  polygonLoader:addTerrainFromObjectLayer(
    self.collider,
    objs.terrain,
    'level'
  )
  --[[
  -- XXX: Add actual collision mesh from polygons here
  buildCollisionMesh(self.collider, self.tilemap.width, self.tilemap.height, self.tilemap.tilewidth, self.tilemap.tileheight,
    function(x, y)
      return self.mainLayer:get_tile(x-1, y-1)
    end,
    function(a)
      return a == 0
    end)
  ]]


  --self.mainLayerRenderer = coup.tilelayerrenderer.TilelayerRenderer(self.mainLayer, 10)

--  self.bgLayer = self.tilemap:get_layer("background")
--  self.bgLayerRenderer = coup.tilelayerrenderer.TilelayerRenderer(self.bgLayer, 10)
--  self.overlayLayer = self.tilemap:get_layer("overlay")
--  self.overlayLayerRenderer = coup.tilelayerrenderer.TilelayerRenderer(self.overlayLayer, 10)

  self.monster = createMonster({
    speed = 50,
    hidden = true,
    paused = true,
    paths = objs.eyepath.objects,
    avoidance = 350,
  })

  self.sparks = Sparks()
  self.dust = Dust()

  self.dead_tween = coup.tween.group()
  self.tween = coup.tween.group()
  self.message_tween = coup.tween.group()
  self.world = coup.world.World()
  coup.components.register(self.world)
  self.world:registerComponentType("PhysicsObject", require("physics_object")())
  self.world:registerComponentType("Input", require("input")())
  self.world:registerComponentType("Sensors", require("sensors")())
  self.world:registerComponentType("Highlight", require("highlight")())
  
  self:createObjects(objs)

  self.camera = self.world:newEntity("camera")
  local w, h = G.screenSize[1], G.screenSize[2]
  self.camera:makeComponent("Transform", w/2, h/2)
  self.camera:makeComponent("BoundingBox", -w/2, -h/2-180, w/2, h/2-180)

  self.spriteRenderer     = SpriteRenderer(self.collider)
  self.spriteUpdater      = coup.systems.SpriteUpdater(self.world)
  self.physicsUpdater     = PhysicsUpdater(self)
  self.highlightRenderer  = HighlightRenderer(self.collider)

  self.playerController = PlayerController(self.player)
  self.playerProcessor = PlayerProcessor(self)

  self.world:step()

  self.vignette = G.loadImage("data/vignette.png")
  self.eyesImage = G.loadImage('data/eyes.png')
  self.lightCanvas = love.graphics.newCanvas(w, h)

  self.timerTween = {t = 0}
  --[[
  self.tween:tween(self.timerTween, 0.5, {t = 1}):on_complete(function()
    self:show_message("Welcome, friend!\nMy plan is to create the philosopher's stone today. To do that, I will have to find all four elemental orbs.", function()
      self:show_message("Each of the orbs will also allow me to brew a potion that will give me new abilities. Those should come in handy to find the remaining ones!", function()
        self:show_message("Walk around with [LEFT] and [RIGHT] or your [DPAD]. Jump with [X] on your keyboard or [A] on your gamepad.")
      end)
    end)
  end)
  ]]
  

end


function Level:updateCamera(lerp)
  local p = self.camera:getComponent("Transform")
  local pp = self.player:getComponent("Transform")

  if not self.playerDead then
    p.pos = pp:lpos(lerp):clone()
  end
  AABBUpdater_updateEntity(self.camera)

end


function handlePlayerTrigger(level, player, trigger)
  level.monster:handleEvent(trigger.properties)

  -- Ensure trigger only happens once.
  level.collider:setGhost(trigger.shape)
  level.collider:remove(trigger.shape)
end


function Level:show_tut_1()
  self:showMessage("Run: left/right")
end


function Level:show_tut_2()
  self:showMessage("Jump: x")
end

function Level:show_tut_3()
  self:showMessage("Hold shift to walk")
end


function Level:show_freedom()
  self:showMessage("Freedom", function()
    self.victoryEyes = {}
    local p = self.player:getComponent('Transform'):lpos(1)
    function addEye(i)
      i = i + love.math.randomNormal(0.3, 0)
      local x = p.x + 500 * math.cos(i) + math.sin(30 * i) + love.math.randomNormal(10, 0)
      local y = p.y + 400 * math.sin(i) + math.cos(30 * i) + love.math.randomNormal(30, 0)
      table.insert(self.victoryEyes, {x, y, love.math.random()})
    end
    for j=0,3 do
      addEye(j/4-3)
    end
    for j=7,10 do
      addEye(j/4-3)
    end
    self.victory = {fear = 0, vis = 0, heart=1}
    self.tween:tween(self.victory, 5, {fear = 300, vis=1})
      :on_update(function() 
          for i, amb in pairs(G.ambient) do
            amb:setVolume((1-self.victory.vis) * 0.3)
          end
        end)
      :after(self.victory, 3, {})
      :on_complete(function()
          self.victory.black = 0
          self.tween:tween(self.victory, 5, {black = 1})
            :after(self.victory, 3, {heart=0})
            :on_complete(function()
              self:showMessage("End")
            end)
        end)

  end)
end


function handleLevelTrigger(level, player, trigger)
  -- Ensure trigger only happens once.
  --debugging:print(trigger.properties.func)
  level[trigger.properties.func](level)
  level.collider:setGhost(trigger.shape)
  level.collider:remove(trigger.shape)
end

local collisionEvents = {
  {"player", "trigger", handlePlayerTrigger},
  {"player", "leveltrigger", handleLevelTrigger}
}


function Level:handleCollision(a, b)
  for i, c in ipairs(collisionEvents) do
    if a.item == c[1] and b.item == c[2] then
      c[3](self, a, b)
    elseif a.item == c[2] and b.item == c[1] then
      c[3](self, b, a)
    end
  end
end






local timestep = 1 / 200
function Level:update(dt)
 
  if self.heart2_time then
    if self.heart2_time < self.totalTime then
      self.heart2_time = nil
      G.heartbeat2:play()
    end
  end
  if self.player.fear > 0 then
    local bpm = (130/400) * self.player.fear + 70
    local spb = 60 / bpm
    if self.totalTime - self.player.last_heartbeat > spb then
      local volume = self.player.fear / 200
      if self.victory then
        volume = self.victory.heart
      end
      G.heartbeat:setVolume(volume)
      G.heartbeat2:setVolume(volume*0.8)
      self.heart2_time = self.totalTime + spb / 3
      G.heartbeat:play()
      self.player.last_heartbeat = self.totalTime
    end
  end

  if self.player.dead then
    self.dead_tween:update(dt)
    self.sparks:update(dt)
    self.totalTime = self.totalTime + dt
    return
  end

  self.t = self.t + dt

  local linearStep = 0
  while self.t > timestep do
    self.playerController:update()
    if self.victoryEyes then
      local input = self.player:getComponent('Input')
      input.left = false
      input.right = false
      input.jump = false
    end
    if not self.playerDead then
      self.playerProcessor:update()
    end
    self.physicsUpdater:update(timestep)
    for i, c in ipairs(self.physicsUpdater.collisions) do
      self:handleCollision(c[2], c[3])
    end
    linearStep = linearStep + timestep
    self.t = self.t - timestep
  end
  if self.player.torch_on then
    self.sparks:emit(function(l) return self:getTorchPos(l) end, linearStep)
  end

  local playerTransform = self.player:getComponent('Transform')
  local spd = math.abs(self.player:getComponent('Velocity').vel.x)
  if spd > 1 then
    local sen = self.player:getComponent("Sensors")
    if sen.sensors.foot:touchesGround(self.collider) then
      self.dust:emit(spd / 4, function(l)
          local p = playerTransform:lpos(l):clone()
          return {p.x, p.y}
        end, linearStep)
    end
  end

  self.totalTime = self.totalTime + linearStep
  self:updateCamera(self.t / timestep)
  self.spriteUpdater:updateSprites(linearStep)
  local avoidance = math.sqrt(
    coup.smath.sq(playerTransform.pos.x - self.monster.pos[1]) +
    coup.smath.sq(playerTransform.pos.y - self.monster.pos[2])
  )
  if avoidance > self.monster.avoidance then
    self.monster:update(linearStep)
  end
  self.sparks:update(linearStep)
  self.dust:update(linearStep)
  self.tween:update(linearStep)
  self.message_tween:update(linearStep)
  self:update_fear(linearStep)

  local spr = self.player:getComponent('Sprite')
  local frame = spr.sprite:getCurrentFrame().index
  local anim_name = spr.sprite.anim.name

  function crossed(x)
    return frame >= x and (spr.last_frame < x or spr.last_frame > frame)
  end

  if spr.last_frame ~= frame then
    if (anim_name == 'Run' and (crossed(3) or crossed(19)))
      or (anim_name == 'Walk' and (crossed(1) or crossed(33))) then
        G.step:setVolume((anim_name == 'Run') and 0.25 or 0.125)
        G.step:play()
    end
  end


  spr.last_frame = frame

end

function math.isnan(value)
  if type(value) == "string" then
    value = tonumber(value)
    if value == nil then return nil end
  elseif type(value) ~= "number" then
    return nil
  end
  return value ~= value
end


function Level:kill(reason)
  if reason == "drop" then
    G.drop:play()
    self.player.fear = 0
    for i,amb in ipairs(G.ambient) do
      amb:stop()
    end
    self.monster_wait = {t=0}
    self.dead_state = "show_message"
    self.dead_tween:tween(self.monster_wait, 2, {t=2}):on_complete(function()
    end)

  elseif reason == "monster" then
    self.monster_wait = {t=0}
    local monster_start_pos = {self.monster.pos[1], self.monster.pos[2]}
    local player_pos = self.player:getComponent("Transform"):lpos(1):clone()
    self.dead_state = "monster_wait"
    self.dead_tween:tween(self.monster_wait, 1, {t=1}):on_complete(function()
      G.pounce:play()
      for i,amb in ipairs(G.ambient) do
        amb:stop()
      end
      self.monster_wait.t=0
      self.dead_state="monster_attack"
      local jump_length = 1.0
      local jump_height = 30
      local fa = -4*(jump_height / (jump_length*jump_length))
      local fb = 4*(jump_height / jump_length)
      self.dead_tween:tween(self.monster_wait, 0.2, {t=0.6})
        :ease('sine_in')
        :on_update(function()
          local offset = fa*self.monster_wait.t*self.monster_wait.t + fb*self.monster_wait.t
          self.monster.pos[1] = coup.smath.lerp(monster_start_pos[1], player_pos.x,    self.monster_wait.t)
          self.monster.pos[2] = coup.smath.lerp(monster_start_pos[2], player_pos.y-40, self.monster_wait.t) - offset
        end)
        :on_complete(function()
          self.dead_state="monster_hit"
          self.player.fear = 0
          self.monster_wait.t=0
          self.dead_tween:tween(self.monster_wait, 0.1, {t=1}):on_complete(function()
            self.dead_state="streak_fade"
            self.monster_wait.t=0
              self.dead_tween:tween(self.monster_wait, 1, {t=1}):on_complete(function()
                self.dead_state="show_message"
                self.monster_wait.t=0
                self.dead_tween:tween(self.monster_wait, 2, {t=2}):on_complete(function()
                end)
              end)
          end)
        end)
    end)
  end
  self.player.dead = reason
end


function Level:update_fear(dt)
  function monster_fear()
    local mx, my = self.monster.pos[1], self.monster.pos[2]
    local pp = self.player:getComponent('Transform'):lpos(1):clone()
    local px, py = pp.x, pp.y
    local dx, dy = px-mx, py-my
    local dist = math.sqrt(dx*dx+dy*dy)
    if not math.isnan(dist) then
      dist = math.max(dist, eye_strike_distance)

      if dist <= eye_strike_distance then
        self:kill("monster")
      end

      return fear_from_distance(dist)
    end
  end

  function drop_fear()
    local player = self.player
    local sensors = player:getComponent('Sensors')
    local v = player:getComponent('Velocity')
    local onGround = sensors.sensors.foot:touchesGround(self.collider) or v.vel.y < 150
    if onGround then
      player.last_contact = self.totalTime
      return 0
    else
      local drop_time = self.totalTime - player.last_contact - min_drop_fear
      if drop_time > max_drop_fear  then
        self:kill('drop')
        return 0
      end
      return 400 * math.max(0, math.min(1, drop_time / (max_drop_fear - min_drop_fear)))
    end
  end

  local fear_m = monster_fear()
  local fear_d = drop_fear()
  local fear = math.max(fear_m, fear_d)
  if self.victoryEyes then
    fear = self.victory.fear
  end
  if fear > 0 then
    self.player.fear = math.max(self.player.fear, fear)
    self.player.fear_timeout = fear_timeout
  end



  self.player.fear_timeout = self.player.fear_timeout - dt
  if self.player.fear_timeout <= 0 then
    self.player.fear = self.player.fear - fear_decay_rate * dt
    if self.player.fear < 0 then
      self.player.fear = 0
    end
    self.player.torch_on = self.player.fear_timeout < fear_timeout/2
  else
    if self.player.fear < min_flicker_fear then
      self.player.torch_on = true
    else
      local prob = math.min(1, math.max(0, (self.player.fear - eye_strike_fear) / (min_flicker_fear - eye_strike_fear)))

      if self.player.torch_on then
        self.player.torch_on = self.player.torch_on and math.random() < prob
      else
        self.player.torch_on = self.player.torch_on or math.random() < prob
      end
    end
  end
end

function Level:show_message(message, next_func)
  --[[
  self.message = {
    text = message,
    size = 0,
    next_func = next_func
  }

  self.tween:tween(self.message, 0.3, {size = 1})
  --]]
end


function Level:debug_transform()
  self.world:foreachEntityWithComponents(function(e)
    local t = e:getComponent("Transform")
    love.graphics.line(t.pos.x - 4, t.pos.y - 4, t.pos.x + 4, t.pos.y + 4)
    love.graphics.line(t.pos.x + 4, t.pos.y - 4, t.pos.x - 4, t.pos.y + 4)
  end, "Transform")
end

function Level:debug_entity_names()
  self.world:foreachEntityWithComponents(function(e)
    local t = e:getComponent("Transform")
    love.graphics.print(e.name, t.pos.x + 5, t.pos.y)
  end, "Transform")
end

function Level:draw()
  love.graphics.setColor(1,1,1,1)
  local cab = self.camera:getComponent("AxisAlignedBoundingBox")
  if self.player.dead then
    love.graphics.setBlendMode('alpha')
    love.graphics.setColor(1,1,1,1)
    if self.dead_state == "monster_wait" or self.dead_state == "monster_attack" then
      love.graphics.clear(0,0,0,1)
      self:inScreenSpace(cab.topLeft, cab.bottomRight, function()
        self:drawMonster()
        self.sparks:draw()
      end)
    elseif self.dead_state == "monster_hit" or self.dead_state == "streak_fade" then
      if self.dead_state == "monster_hit" then
        love.graphics.clear(math.max(0, 1-self.monster_wait.t),0,0,1)
      else
        love.graphics.clear(0,0,0,1)
      end
      local w, h = G.streak:getDimensions()
      local b = 1
      if self.dead_state == "streak_fade" then
        b = 1-self.monster_wait.t
      end
      love.graphics.setColor(b,b,b,1)
      love.graphics.draw(G.streak, (G.screenSize[1]-w)/2, 0)
      if self.dead_state == "monster_hit" then
        love.graphics.setColor(math.max(0, 1-self.monster_wait.t),0,0,1)
        love.graphics.rectangle('fill', 0, G.screenSize[2]*self.monster_wait.t, G.screenSize[1], G.screenSize[2])
      end
    elseif self.dead_state == "show_message" then
      love.graphics.setColor(1,0,0,math.max(0,self.monster_wait.t-1))
      love.graphics.printf("PRESS X TO RETRY", 0, G.screenSize[2]*0.8, G.screenSize[1], 'center')
    end

    love.graphics.setColor(1,1,1,1)
    

    return
  end

  love.graphics.clear(50/255, 40/255, 80/255, 255/255)

  local lerp = self.t / timestep
  --self.bgLayerRenderer:draw(cab.topLeft, cab.bottomRight)

  love.graphics.setBlendMode('add')
  self.highlightRenderer:renderVisible(self.camera, lerp, {64/255,64/255,64/255,255/255}, true)
  love.graphics.setBlendMode('alpha')
  love.graphics.setColor(1,1,1,1)

  self:inScreenSpace(cab.topLeft, cab.bottomRight, function()
    self:drawTerrain()
  end)
  --self.mainLayerRenderer:draw(cab.topLeft, cab.bottomRight)
  self.spriteRenderer:renderVisible(self.camera, lerp)

  --self.overlayLayerRenderer:draw(cab.topLeft, cab.bottomRight)



  local can = love.graphics.getCanvas()
  love.graphics.setCanvas(self.lightCanvas)
  love.graphics.clear(0,0,0,1)

  love.graphics.setBlendMode('add')
  --love.graphics.setColor(1,1,1,1)
  --love.graphics.draw(self.vignette, 0, 0, 0, 2, 2)
  --self.highlightRenderer:renderVisible(self.camera, lerp)

  love.graphics.setCanvas(can)

  self:drawTorch(lerp, cab, self.player.fear, self.player.torch_on)

  self:inScreenSpace(cab.topLeft, cab.bottomRight, function()
    self.sparks:draw()
    self.dust:draw()
  end)

  love.graphics.setBlendMode('multiply', 'premultiplied')
  --love.graphics.setBlendMode('darken', 'premultiplied')
  love.graphics.draw(self.lightCanvas)

  love.graphics.setBlendMode('alpha')

  self:inScreenSpace(cab.topLeft, cab.bottomRight, function()
    self:drawMonster()
    self.sparks:draw()
    if self.victoryEyes then
      love.graphics.setBlendMode('alpha')
      love.graphics.setColor(1,1,1,1)
      for i, eyePos in ipairs(self.victoryEyes) do
        if eyePos[3] < self.victory.vis then
          self:drawMonster(eyePos)
        end
      end
    end
  end)

  if self.victory then
    if self.victory.black then
      love.graphics.setColor(0,0,0,self.victory.black)
      love.graphics.setBlendMode('alpha')
      love.graphics.rectangle('fill', 0, 0, G.screenSize[1], G.screenSize[2])
      love.graphics.setColor(1,1,1,1)
    end
  end

  self:drawMessage()

--  self:debug_transform()
--  self:debug_entity_names()
end


function Level:getTorchPos(lerp)
  lerp = lerp or 0
  local ps, pt = self.player:getComponents('Sprite', 'Transform')
  local pos = pt:lpos(lerp)
  local anim_name = ps.sprite.anim.name
  local frame = ps.sprite:getCurrentFrame()
  local frame_idx = frame.index
  local spr_torch_pos = G.torches[anim_name][frame_idx]
  return {(spr_torch_pos[1]-64) * pt.scale.x + frame.origin.x + pos.x, (spr_torch_pos[2]-128) * pt.scale.y + frame.origin.y + pos.y}
end


function Level:drawTorch(lerp, camera_bounding_box, fear, torch_on)
  if not torch_on then
    return
  end

  local world_torch_pos = self:getTorchPos()
  love.graphics.push()
  local cameraSize, screenOffset, offset = coup.camera.cameraSettings( camera_bounding_box.topLeft, camera_bounding_box.bottomRight)
  love.graphics.translate(math.floor(offset.x), math.floor(offset.y))

--  if torch_on then
  love.graphics.setBlendMode('add')
  local bright = love.math.randomNormal(0.1, 0.6)
  love.graphics.setColor(bright, bright, bright, 1.0)
  love.graphics.draw(G.torch, world_torch_pos[1], world_torch_pos[2], 0, love.math.randomNormal(0.1, 0.7), love.math.randomNormal(0.1, 0.7), 40, 40)
  love.graphics.setColor(0.5,0.5,0.5,1)
  love.graphics.draw(G.torch_glow, world_torch_pos[1], world_torch_pos[2], 0, 1+0.1*math.sin(self.t), 1+0.1*math.sin(self.t), 40, 40)



  -- do light
  local can = love.graphics.getCanvas()
  love.graphics.setCanvas(self.lightCanvas)
  love.graphics.setBlendMode('replace')
  --bright = love.math.randomNormal(0.1, 0.9)
  local rot = love.math.random() * 2 * 3.14159
  local scale = love.math.randomNormal(0.01, 1)
  bright = 1
  love.graphics.setColor(bright, bright, bright, 1.0)
  scale = scale * (torch_range - fear) / torch_range
  love.graphics.draw(G.torch_light, world_torch_pos[1], world_torch_pos[2], rot, scale, scale, 400, 400)

  local shapes = self.collider:shapesInRange(world_torch_pos[1] - torch_range, world_torch_pos[2] - torch_range, world_torch_pos[1] + torch_range, world_torch_pos[2] + torch_range)

  local shapes_array = {}
  for k,v in pairs(shapes) do
    if v.islevel then
      shapes_array[#shapes_array+1] = v
    end
  end
  

  love.graphics.setColor(0,0,0,1)
  for i = 1, #shapes_array do
    local shape = shapes_array[i]
    local shape_vertices = {shape._polygon:unpack()}
    local front_vertices = {}
    local sc_x, sc_y = shape:center()
    local corner_candidates = {}
    for ei = 1,#shape_vertices,2 do
      local i1 = ei
      local i2 = ((ei+2)-1) % #shape_vertices + 1
      local x1, y1, x2, y2 = shape_vertices[i1], shape_vertices[i1+1], shape_vertices[i2], shape_vertices[i2+1]
      if(same_side(x1,y1,x2,y2, world_torch_pos[1], world_torch_pos[2], sc_x, sc_y)) then
        corner_candidates[i1] = not corner_candidates[i1] or nil
        corner_candidates[i2] = not corner_candidates[i2] or nil
      end
    end

    local actual_corners = {}
    for i, _ in pairs(corner_candidates) do
      actual_corners[#actual_corners+1] = i
    end

    local low = actual_corners[1]
    local high = actual_corners[2]
    if low and high then
      local dx1 = shape_vertices[low] - world_torch_pos[1]
      local dy1 = shape_vertices[low+1] - world_torch_pos[2]
      local dx2 = shape_vertices[high] - world_torch_pos[1]
      local dy2 = shape_vertices[high+1] - world_torch_pos[2]
      local s1 = math.sqrt(dx1*dx1+dy1*dy1)
      local s2 = math.sqrt(dx2*dx2+dy2*dy2)
      dx1 = dx1 * s1
      dy1 = dy1 * s1
      dx2 = dx2 * s2
      dy2 = dy2 * s2

      ---[[
      love.graphics.polygon('fill', shape_vertices[low], shape_vertices[low+1],
                         shape_vertices[high], shape_vertices[high+1],
                         shape_vertices[high] + 2*torch_range*dx2, shape_vertices[high+1]+2*torch_range*dy2,
                         shape_vertices[low] + 2*torch_range*dx1, shape_vertices[low+1] + 2*torch_range*dy1
                         )
      --]]
    end
  end
  love.graphics.setCanvas(can)
  love.graphics.setColor(1,1,1,1)

  love.graphics.pop()
end

function same_side(x1, y1, x2, y2, ax, ay, bx, by)
  return  ((y1 - y2)*(ax - x1) + (x2 - x1)*(ay - y1))
        * ((y1 - y2)*(bx - x1) + (x2 - x1)*(by - y1)) < 0
end


function Level:showMessage(text, onComplete)
  self.message = {text = text, a = 0}
  self.message_tween:abort_all()
  self.message_tween:tween(self.message, 0.5, {a = 0.8})
    :after(self.message, 5, {a = 0.8})
    :after(self.message, 2, {a = 0})
    :on_complete(function()
      self.message = nil
      if onComplete then
        onComplete()
      end
    end)
end


function Level:inScreenSpace(topLeft, bottomRight, func)
  local x1, y1 = topLeft.x, topLeft.y
  local x2, y2 = bottomRight.x, bottomRight.y
  local g = love.graphics
  g.push()
  local cameraSize, screenOffset, offset = coup.camera.cameraSettings(
    coup.vector.Vector(x1, y1),
    coup.vector.Vector(x2, y2)
  )
  g.translate(math.floor(offset.x), math.floor(offset.y))
  func()
  g.pop()
end


function Level:drawMonster(pos)
  if not pos then pos = self.monster.pos end
  if not self.monster.hidden then
    local scaleX = 1
    local playerPosition = self.player:getComponent('Transform').pos
    if pos[1] > playerPosition.x then
      scaleX = -1
    end
    love.graphics.draw(
      self.eyesImage,
      pos[1], pos[2],
      self.monster:angleTowards(playerPosition, pos),
      scaleX, 1,
      self.eyesImage:getPixelWidth() / 2, self.eyesImage:getPixelHeight() / 2
    )
  end
end


function Level:drawTerrain()
  local phys = self.camera:getComponent("AxisAlignedBoundingBox")
  local tl = phys.topLeft
  local br = phys.bottomRight
  local shapes = self.collider:shapesInRange(tl.x - 100, tl.y - 100, br.x + 100, br.y + 100)
  love.graphics.setColor(0,0,0,255)
  for k, shape in pairs(shapes) do
    if shape.islevel then
      shape:draw('fill')
    end
  end
end


function Level:drawMessage()
  --self.message.text = "Foop"
  -- print(self.message)
  if self.message then
    local w = G.screenSize[1]
    local h = G.screenSize[2]
    --[[
    love.graphics.setColor(0,0,0,224)
    love.graphics.setLineWidth(2)
    love.graphics.rectangle('fill', (G.screenSize[1] - w) / 2, G.screenSize[2] * 0.75 - h / 2, w, h)
    love.graphics.setColor(255,255,255,255)
    love.graphics.rectangle('line', (G.screenSize[1] - w) / 2, G.screenSize[2] * 0.75 - h / 2, w, h)

    if self.message.size == 1 then
    ]]
      love.graphics.setColor(1,1,1,self.message.a)
      love.graphics.printf(self.message.text,
        0, h * 0.85, w, "center")
    --[[
      love.graphics.setColor(128,128,128,255)
      love.graphics.printf("Hit [RETURN] or [B] on Gamepad", 
        (G.screenSize[1] - w) / 2 + 5,
         G.screenSize[2] * 0.75 + h / 2 - 10,
         w,
         "center")
    end
     ]]
      love.graphics.setColor(1,1,1,1)


  end
end

function Level:play_roar()
  G.pounce:play()
end



return Level
